@@ -16,7 +16,7 @@
 import os
 import sys
 import warnings
-from pkg_resources import parse_version
+from unittest import mock
 
 import setuptools  # force import setuptools before any other distutils imports
 
@@ -25,24 +25,14 @@ from buildbot.test.util.warnings import assertProducesWarning  # noqa pylint: di
 from buildbot.test.util.warnings import assertProducesWarnings  # noqa pylint: disable=wrong-import-position
 from buildbot.warnings import DeprecatedApiWarning  # noqa pylint: disable=wrong-import-position
 
-# import mock so we bail out early if it's not installed
-try:
-    import mock
-    [mock]
-except ImportError as e:
-    raise ImportError("\nBuildbot tests require the 'mock' module; "
-                      "try 'pip install mock'") from e
+[mock]
 
 # apply the same patches the buildmaster does when it starts
-monkeypatches.patch_all(for_tests=True)
+monkeypatches.patch_all()
 
 # enable deprecation warnings
 warnings.filterwarnings('always', category=DeprecationWarning)
 
-if parse_version(mock.__version__) < parse_version("0.8"):
-    raise ImportError("\nBuildbot tests require mock version 0.8.0 or "
-                      "higher; try 'pip install -U mock'")
-
 [setuptools]  # force use for pylint
 
 # This is where we load deprecated module-level APIs to ignore warning produced by importing them.
@@ -87,10 +77,6 @@ warnings.filterwarnings('ignore', r"The value of convert_charrefs will become Tr
 # Twisted 18.4+ adds a deprecation warning and still use the deprecated API in its own code!
 warnings.filterwarnings('ignore', ".*getClientIP was deprecated.*", DeprecationWarning)
 
-# Python 3.7 adds a deprecation warning formatargspec.
-# The signature api that replaces it is not available in 2.7
-warnings.filterwarnings('ignore', ".*`formatargspec` is deprecated.*", DeprecationWarning)
-
 # Python 3.7 adds a deprecation importing ABCs from collection.
 # Such imports are made in dependencies (e.g moto, werzeug, pyparsing)
 warnings.filterwarnings('ignore', ".*Using or importing the ABCs from 'collections'.*",
@@ -143,3 +129,21 @@ warnings.filterwarnings('ignore', ".*stream argument is deprecated. Use stream p
 # use. See https://github.com/boto/botocore/issues/2744
 warnings.filterwarnings('ignore', ".*'urllib3.contrib.pyopenssl' module is deprecated",
                         category=DeprecationWarning)
+
+# pipes is still used in astroid and buildbot_worker in default installation
+warnings.filterwarnings('ignore', "'pipes' is deprecated and slated for removal in Python 3.13",
+                        category=DeprecationWarning)
+
+# shown on Python 3.7 on Windows
+warnings.filterwarnings('ignore', "SelectableGroups dict interface is deprecated. Use select.",
+                        category=DeprecationWarning)
+
+# shown on Python 3.11
+warnings.filterwarnings('ignore', ".*pkg_resources is deprecated as an API.*",
+                        category=DeprecationWarning)
+warnings.filterwarnings('ignore', ".*Deprecated call to `pkg_resources.declare_namespace.*",
+                        category=DeprecationWarning)
+
+# boto3 shows this warning when on old Python
+warnings.filterwarnings('ignore', ".*Boto3 will no longer support Python .*",
+                        category=Warning)
